/*
   penbw - pen backend watcher
   Copyright (C) 2006 Thomas A.F. Eckhardt <tafe unkelhaeusser net>
   based on and for pen by Ulric Eriksson <ulric siag nu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

//#define VERSION "0.4"
#define EXE "penbw"
#define AUTHOR "Thomas A.F. Eckhardt <tafe unkelhaeusser net>"

#define CHK_REQUEST "HEAD / HTTP/1.0\r\n\r\n"
//#define CHK_REQUEST "GET / HTTP/1.0\r\n\r\n"

#define DEFAULT_TIMEOUT 5
#define LAZY_MAXSRV 20
#define BUFFER_SIZE 128
#define SLP_AFTCHK 2
#define PSV_SSKSZ 40
#define PSV_SPSZ 5

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

#include <netdb.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <errno.h>
extern int errno;

#include <syslog.h>
#include <stdarg.h>

#define IFDEBUG(dbglvl) if(debug > dbglvl)

struct psv{
    short int sid;
    char ssk[PSV_SSKSZ];
    char sp[PSV_SPSZ];
};
// this should really be dynamic :(
struct psv svlst[LAZY_MAXSRV];

static short int sleepdelay = 60;
static struct timeval tv;
static int foreground = 0;
static short int debug = 0;
static short int dummy = 0;
static char *unpriv_user;

void tlog(int ltype, char *fmt, ...) {
	
	char buffer[1024];

	va_list arglist;
	va_start(arglist, fmt);
	vsnprintf(buffer, sizeof(buffer), fmt, arglist);
	va_end(arglist);

	if (foreground != 1) {
		openlog(EXE, LOG_CONS | LOG_PID, LOG_USER);
		syslog(ltype, "%s\n", buffer);
		closelog();
	} else {
		if (ltype == LOG_DEBUG || ltype == LOG_INFO || ltype == LOG_NOTICE) {
			printf("%s\n", buffer);
		} else { fprintf(stderr, "%s\n", buffer); } 
	}
}

unsigned int drop_privileges(char *tuname) {

	struct passwd *ud = NULL;

	if (getuid() == 0) {
		ud = getpwnam(tuname);
		if (ud == NULL) {
			tlog(LOG_ERR, "Unknown user: %s", tuname);
			exit(1);
		} else { 
			if ( (setgid(ud->pw_gid) == -1) || (setuid(ud->pw_uid) == -1) )
				{
					tlog(LOG_ERR, "Can't drop privilegs to user %s: %s", tuname, strerror(errno));
					exit(1);
				}
		}
	} else { tlog(LOG_WARNING, "Can't drop privileges as unprivileged user"); return 1; }

	return 0;
}

void usage(void)
{
	printf("%s v%s by %s\n", EXE, VERSION, AUTHOR);
        printf("Usage: penbw [OPTIONS] PENCTLHOST:PENCTLPORT\n");
        printf("\n"
               "  -u username\tswitch as privileged user to unprivileged user after\n"
	       "             \tstartup and basic sanity checks\n"
               "  -s sec\ttime to sleep after each check-round [%i]\n"
               "  -t sec\tconnect/send/receive timeout to backends [%i]\n"
	       "  -f    \tstay in foreground [off]\n"
	       "  -n    \tdon't blocklist, just watch [off]\n"
	       "  -d    \tincrease debug lvl (repeatable) [off]\n"
               "  -h, --help\n"
               "  -V, --version\n"
               "\n", sleepdelay , DEFAULT_TIMEOUT);

	puts("example:\n  pwnbw -s 30 127.0.0.1:1080\n");
	exit(0);
}

static int options(int argc, char **argv)
{
        int c;
        char *opt = "s:t:u:Vhfdn";

        while ((c = getopt(argc, argv, opt)) != -1) {
                switch (c) {
                case 'u':
			unpriv_user = optarg;
                        break;

                case 'f':
			foreground = 1;
                        break;

                case 'd':
			++debug;
                        break;

                case 'n':
			dummy = 1;
                        break;

                case 's':
			sleepdelay = atoi(optarg);
                        break;

                case 't':
			tv.tv_sec = atoi(optarg);
                        break;

                case 'V':
			printf("%s v%s - Pen Backend Watcher\n", EXE, VERSION);
			exit(0);
                        break;

		case 'h':
                case '?':
                default:
                        usage();
                }
        }

	// init tv
	tv.tv_sec = DEFAULT_TIMEOUT;
	tv.tv_usec = 0;


        return optind;
}

void daemonize(void) {
        int child;
        if ((child = fork()) < 0) {
                perror("fork() failed");
		exit(1);
        } else if (child == 0) {
		// we are the child
		setsid();
	} else {
		// we are the parent
		exit(0);
        }
}

// TODO: cleanup
unsigned int open_socket(char *addr, char *port)
{
        int fd = -1;
        struct sockaddr_in serv_addr;
        struct hostent *h;
        struct in_addr a;
        struct servent *s;
        int n, po;
        h = gethostbyname(addr);
        if (h == NULL) {
		IFDEBUG(1) tlog(LOG_DEBUG, "unknown or invalid address %s", addr);
		return -1;
	}
        memcpy(&a, h->h_addr, h->h_length);
        s = getservbyname(port, "tcp");
        if (s == NULL) po = atoi(port);
        else po = ntohs(s->s_port);

        fd = socket(AF_INET, SOCK_STREAM, 0);
        if (fd < 0) {
		IFDEBUG(1) perror("Error opening socket()");
		return -1;
	}
        serv_addr.sin_family = AF_INET;
        serv_addr.sin_addr.s_addr = a.s_addr;
        serv_addr.sin_port = htons(po);
//	signal(SIGALRM, SIG_IGN);
//	alarm(tv.tv_sec);
        n = connect(fd, (struct sockaddr *)&serv_addr, sizeof serv_addr);
//	alarm(0);
        if (n == -1) {
		IFDEBUG(1) tlog(LOG_DEBUG, "connect() to %s:%s failed: %s", addr, port, strerror(errno));
		return -1;
        }

        return fd;
}

unsigned int blacklist_backend(int penid, char *penctlsk, char *penctlp) {

	int fd, blocktime;
	char penblreq[75];

	// ;)
	blocktime = (sleepdelay + (SLP_AFTCHK * LAZY_MAXSRV)) * 1.75;
	sprintf(penblreq, "server %i blacklist %i", penid, blocktime);

	fd = open_socket(penctlsk, penctlp);
        if (fd < 0) {
                IFDEBUG(1) tlog(LOG_DEBUG, "Socket failed: %s", strerror(errno));
                return 1;
        }
        setsockopt(fd,SOL_SOCKET,SO_SNDTIMEO,&tv,sizeof(tv));
	
        if (send(fd, penblreq, strlen(penblreq), 0) != strlen(penblreq)) {
                IFDEBUG(1) tlog(LOG_DEBUG, "couldn't blacklist, send() to pen incomplete: %s", strerror(errno));
                return 1;
        }

	// TODO: check for response of pen

	close (fd);

	tlog(LOG_NOTICE, "server id(%i) blacklisted (%ds)", penid, blocktime);

	return 0;
}

// update array of available pen-servers to check
// TODO: cleanup
int penctl_readservers(char *penctlsk, char *penctlp) {

	int fd, ss = 0, rc, tsid;
	char *lToken;

	char buffer[BUFFER_SIZE], c, *bufferpnt;

	fd = open_socket(penctlsk, penctlp);
	// ^- TODO!!
	if (fd == -1) { tlog(LOG_ERR, "error connecting to pen at %s:%s", penctlsk, penctlp); return 1; }

	write(fd, "servers\n", strlen("servers\n"));

	bzero(buffer, BUFFER_SIZE);
	bufferpnt = buffer;

	while ((rc = read(fd, &c, 1)) > 0) {
	  IFDEBUG(4) tlog(LOG_DEBUG, "%c", c);
	  *bufferpnt++ = c;

	  if (c == '\n') { // got one row
		lToken = strtok(buffer, " ");
		if (lToken)
		  {
			if (atoi(lToken) != 0 || strcmp(lToken, "0") == 0)
			  {
				tsid = atoi(lToken);
				while ( (lToken = strtok(NULL, " ")) )
				  {
					if (strcmp(lToken, "addr") == 0) {
						lToken = strtok(NULL, " ");
						if ( (strcmp(lToken, "0.0.0.0") != 0) && (strlen(lToken) <= 40) ) {
							IFDEBUG(2) tlog(LOG_DEBUG, "Server-Addr: %d -> %s", tsid, lToken);
							//if (ss == (sizeof(servers) / (sizeof(char *)*2))) { fprintf(stderr, "Cannot support More backends! (%d)\n", ss); break; }
							//printf("SS: %d\n", ss);
							svlst[ss].sid = tsid;
							strcpy(svlst[ss].ssk, lToken);
							++ss;
							lToken = strtok(NULL, " ");
							if (strcmp(lToken, "port") == 0) {
								// todo: port looks good?
								lToken = strtok(NULL, " ");
								--ss;
								IFDEBUG(2) tlog(LOG_DEBUG, "Port: %s", lToken);
								if (strlen(lToken) <= 5) {
								//svlst[ss].sp = lToken;
								strcpy(svlst[ss].sp, lToken);
								}
								++ss;
								break;
							}
						}
					}
			  	  }
			  } else { tlog(LOG_WARNING, "Bogus output of penctl request: '%s'", buffer); }
		  } else { tlog(LOG_WARNING, "No suitable output of penctl: '%s'", buffer); }

		//printf("%s", buffer);
		bzero(buffer, BUFFER_SIZE);
		bufferpnt = buffer;
	  }
	}

	if (rc == -1) { tlog(LOG_ERR, "error reading from penctl"); exit(1); }

        close(fd);	

	return ss;
}


unsigned int check_backend(unsigned int idb) {

	int bytes = 0, fd;
        char buffer[BUFFER_SIZE], c, *bufferpnt;
        bzero(buffer, BUFFER_SIZE);
        bufferpnt = buffer;

        fd = open_socket(svlst[idb].ssk, svlst[idb].sp);
	if (fd < 0) {
		IFDEBUG(1) tlog(LOG_WARNING, "Socket failed: %s",strerror(errno));
		return 1;
	}

	setsockopt(fd,SOL_SOCKET,SO_SNDTIMEO,&tv,sizeof(tv));
	if (send(fd, CHK_REQUEST, strlen(CHK_REQUEST), 0) != strlen(CHK_REQUEST)) {
		IFDEBUG(1) tlog(LOG_WARNING, "send did not complete: %s",strerror(errno));
		return 1;
	}

	setsockopt(fd,SOL_SOCKET,SO_RCVTIMEO,&tv,sizeof(tv));
	// we only want the first line
	while (recv(fd, &c, 1, 0) > 0) {
	//while (read(fd, &c, 1) > 0) {
          if (c == '\n' || c == '\r') {
		break;
	  } else {
		*bufferpnt++ = c;
		bytes += 1;
	  }
	}

        if (bytes == 0 || bytes == -1) { return 1; }
	else if (strcmp(buffer, "HTTP/1.1 200 OK") && strcmp(buffer, "HTTP/1.1 302 Found") &&
		 strcmp(buffer, "HTTP/1.1 307 Moved temporarily") && strcmp(buffer, "HTTP/1.1 301 Moved permanently")
		)
		  {
			IFDEBUG(1) tlog(LOG_WARNING, "First Line Response ('%s') does not match", buffer);
			return 1;
		  }

	close(fd);

	return 0;
}

void penbw_end(int signo) {
	tlog(LOG_NOTICE, "penbw v%s exiting on signal %i", VERSION, signo);
	exit(0);
}

int main(int argc, char **argv)
{
	int n;
	char penctlsk[64], *penctlp;
	unsigned short int idx = 0;

        if (argc < 2) {
                fprintf(stderr, "%s: not enough arguments\n", EXE);
                usage();
		exit(1);
        }

	//int n = options(argc, argv);;
	n = options(argc, argv);
        argc -= n; argv += n;

        if ((strlen(argv[argc-1])+1) > sizeof penctlsk) { fprintf(stderr, "Too long arg: '%s'\n", argv[argc-1]); exit(1); }
	strcpy(penctlsk, argv[argc-1]);
	penctlp = strchr(penctlsk, ':');
	if ( penctlp == NULL ) { fprintf(stderr, "no correct penctrl host:port given (%s)\n", argv[argc-1]); exit(1); }
	*penctlp++ = '\0';

	if (unpriv_user != NULL) {
		if (getpwnam(unpriv_user) == NULL) { fprintf(stderr, "Unknown User %s\n", unpriv_user); exit(1); }
	}

	if (foreground != 1) daemonize();
	if (unpriv_user) drop_privileges(unpriv_user);
	tlog(LOG_NOTICE, "penbw v%s started as uid(%i) with penctl(%s:%s) sleepdelay(%i) timeout(%li)", VERSION, getuid(), penctlsk, penctlp, sleepdelay, tv.tv_sec);
	if (dummy == 1) tlog(LOG_NOTICE, "no blocklisting, dummy mode '-n' activated");

	// ignore SIGHUP
	if (signal(SIGHUP, SIG_IGN) == SIG_ERR) tlog(LOG_WARNING, "hu, can't ignore SIGHUP");

	// catch SIGINT & SIGTERM (TODO problem@killall on some systems)
	if (signal(SIGTERM, penbw_end) == SIG_ERR) tlog(LOG_WARNING, "hu, can't catch SIGTERM");
	if (signal(SIGINT, penbw_end) == SIG_ERR) tlog(LOG_WARNING, "hu, can't catch SIGINT");

	// Main-Loop
	for (;;) {
		if (penctl_readservers(penctlsk, penctlp) > 0) {
			IFDEBUG(0) tlog(LOG_DEBUG, "starting a new loop of checks ..");
			while (strcmp(svlst[idx].ssk, "") != 0) {
				IFDEBUG(1) tlog(LOG_DEBUG, "Checking backend: id(%d) addr(%s) port(%s):", svlst[idx].sid, svlst[idx].ssk, svlst[idx].sp);
				if (!check_backend(idx) == 0) {
					tlog(LOG_WARNING, "FAILED backend: id(%d) addr(%s) port(%s)", svlst[idx].sid, svlst[idx].ssk, svlst[idx].sp);
					if (dummy == 0) { 
					  if (blacklist_backend(svlst[idx].sid, penctlsk, penctlp) != 0) {
						tlog(LOG_WARNING, "unable to blacklist backend: id(%d) addr(%s) port(%s)", svlst[idx].sid, svlst[idx].ssk, svlst[idx].sp);
					  }
					}
				} else { 
					IFDEBUG(0) tlog(LOG_NOTICE, "PASSED backend: id(%d) addr(%s) port(%s)", svlst[idx].sid, svlst[idx].ssk, svlst[idx].sp);
				}
				++idx;
				sleep(SLP_AFTCHK);
			}
		} else { IFDEBUG(0) tlog(LOG_DEBUG, "no servers found to check"); } 

		idx = 0;

		// unsure if this is fine
		// (LAZY_MAXSRV * ( sizeof(int *) + (sizeof(char *) * (PSV_SSKSZ + PSV_SPSZ)) ))
		memset(&svlst, 0, sizeof svlst);

		IFDEBUG(0) tlog(LOG_DEBUG, "sleeping for %i seconds", sleepdelay);
		sleep(sleepdelay);
	}
	
	// we never will get here:
        return 0;
}
